/*
 * rv5c387.c
 *
 * Device driver for Ricoh's Real Time Controller RV5C387A.
 *
 * Copyright (C) 2004 Gary Jennejohn garyj@denx.de
 * Copyright (C) Sony corporation. 2007.
 *
 * Based in part in ds1307.c -
 * (C) Copyright 2001, 2002, 2003
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 * Keith Outwater, keith_outwater@mvis.com`
 * Steven Scholz, steven.scholz@imc-berlin.de
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 *
 * Change log:
 * 20080813: AKA <masateru.akachi@jp.sony.com>
 *			*set default value explicitly to CTRL1, CTRL2 register when called 'date reset'
 * 20080812: AKA <masateru.akachi@jp.sony.com>
 *			*changed register read/write sequence to EEP-ROM like (1-byte address) mode.
 *			*changed 24HOUR_MODE bit setting sequence (mode change is taken certainly with time setting)
 * 20080806: AKA <masateru.akachi@jp.sony.com>
 * 			*fixed bugs on hour counter setting in 12-hour mode.
 * 20071104: AKA <masateru.akachi@jp.sony.com>
 * 			*porting from rtc/rs5c372.c
 *			*redefined registers and bit assignment to adjust to RV5C387
 */

#include <common.h>
#include <command.h>
#include <rtc.h>
#include <i2c.h>

#if defined(CONFIG_RTC_RV5C387A) && (CONFIG_COMMANDS & CFG_CMD_DATE)
/*
 * Reads are always done starting with register 15, which requires some
 * jumping-through-hoops to access the data correctly.	==>> NOW NOT USED
 *
 * Writes are always done starting with register 0.	==>> NOW NOT USED
 */
/*
 * Now, inconveniently, RV5C387 register has been remapped from traditional one like RS5C372 and does not meet the legacy "jumping" scheme
 * especially owing to Bit for 12/24(hour mode setting) in Reg.0E ...
 * In this porting, changed access scheme to RTC using EEP-ROM like read/write sequence with standard (1-byte) address instead.
 * Remember that internal address (not I2C slave addr.) has to be given as 4bit shifted value to the left from Register Number to be set.
 * For more details, please refer Chap.1.2-4 Write Transaction Format and 1.2-5 1) Read Transaction Format in the Richo data sheet.
 */


#if DEBUG
static unsigned int rtc_debug = DEBUG;
#else
#define rtc_debug 0	/* gcc will remove all the debug code for us */
#endif

#ifndef CFG_I2C_RTC_ADDR
#define CFG_I2C_RTC_ADDR 0x32
#endif

#define RV5C387_RAM_SIZE	NUM_REGADR

#define	YEARBIT_19_20		7
#define	MASK_YEARBIT_19_20	(1 << YEARBIT_19_20)
#define	IS_YEARBIT_2000(n)	( ((n) & MASK_YEARBIT_19_20) != 0 )

#define STATUS_XSTP  0x10    /* data invalid because voltage was 0 */

#define	RV5C387_RAMADR_TOP	0
enum REG_NUMBER_tag
{
	REGADR_SEC = RV5C387_RAMADR_TOP,	/* 0x0 */
	REGADR_MIN,		/* 0x1 */
	REGADR_HOUR,	/* 0x2 */
	REGADR_WDAY,	/* 0x3 */
	REGADR_DAY,		/* 0x4 */
	REGADR_CENT_MONTH,	/* 0x5 */
	REGADR_YEAR,	/* 0x6 */
	REGADR_CAL,		/* 0x7 */
	REGADR_AW_MIN,	/* 0x8 */
	REGADR_AW_HOUR,	/* 0x9 */
	REGADR_AW_WDAY,	/* 0xA */
	REGADR_AD_MIN,	/* 0xB */
	REGADR_AD_HOUR,	/* 0xC */
	REGADR_UNUSE0,	/* 0xD */
	REGADR_CTRL1,	/* 0xE */
	REGADR_CTRL2,	/* 0xF */
	NUM_REGADR,
};
#define	REGADDR_SHIFT	4
#define	REGADDR_LEN		1
#define	REGADR_24HOUR_MODE	REGADR_CTRL1
#define	REGADR_XSTP			REGADR_CTRL2

#define	MODEBIT_12_24H		5
#define	MASK_MODEBIT_12_24H	(1 << MODEBIT_12_24H)
#define TWELVE_HOUR_MODE(n) (HOURS_AP(n) == 0)
#define HOURS_AP(n)	(((n) >> MODEBIT_12_24H) & 1)
#define HOURS_12(n)	bcd2bin((n) & 0x1F)
#define HOURS_24(n)	bcd2bin((n) & 0x3F)


static uchar bin2bcd (unsigned int n);
static unsigned bcd2bin (uchar c);

static int setup_done = 0;

static int
rv5c387_readram(char *buf, int len)
{
	int ret;

	/* buf[0]<=Reg.0 : This differs from rv5c387.c */
	ret = i2c_read(CFG_I2C_RTC_ADDR, RV5C387_RAMADR_TOP << REGADDR_SHIFT, REGADDR_LEN, buf, len);
	if (ret != 0) {
		printf("%s: failed to read\n", __FUNCTION__);
		return ret;
	}

	if (buf[REGADR_XSTP] & STATUS_XSTP)
		printf("### Warning: RTC lost power\n");

	return ret;
}

static void
rv5c387_enable(void)
{
	unsigned char buf[RV5C387_RAM_SIZE];
	int ret;

	ret = rv5c387_readram(buf, RV5C387_RAM_SIZE);
	if (ret != 0) {
		printf("%s: failed\n", __FUNCTION__);
		return;
	}

	/* Clock rate selection bit (in Reg.7) is not assigned for the RV5C387A, so we have to do nothing */
	setup_done = 1;

	return;
}

static void
rv5c387_convert_to_time(struct rtc_time *dt, char *buf)
{
	dt->tm_sec = bcd2bin(buf[REGADR_SEC]);
	dt->tm_min = bcd2bin(buf[REGADR_MIN]);

	if (TWELVE_HOUR_MODE(buf[REGADR_24HOUR_MODE]))
	{	/* 12-hour-mode */
		dt->tm_hour = HOURS_12(buf[REGADR_HOUR]);
		if( (dt->tm_hour) == 12 )
		{	/* AM12/PM12 */
			dt->tm_hour = 0;
		}

		if (HOURS_AP(buf[REGADR_HOUR]))
		{	/* PM */
			dt->tm_hour += 12;
		}
	}
	else
	{	/* 24-hour-mode */
		dt->tm_hour = HOURS_24(buf[REGADR_HOUR]);
	}

	dt->tm_mday = bcd2bin(buf[REGADR_DAY]);
	dt->tm_year = bcd2bin(buf[REGADR_YEAR]);
	if( IS_YEARBIT_2000( buf[REGADR_CENT_MONTH] ) )
	{
		buf[REGADR_CENT_MONTH] &= ~MASK_YEARBIT_19_20;
		dt->tm_year += 2000;
	}
	else
	{
		dt->tm_year += 1900;
	}
	dt->tm_mon = bcd2bin(buf[REGADR_CENT_MONTH]);
	/* 0 is Sunday */
	dt->tm_wday = bcd2bin(buf[REGADR_WDAY] & 0x07);
	dt->tm_yday = 0;
	dt->tm_isdst= 0;

	if(rtc_debug > 2) {
		printf("rv5c387_convert_to_time: year = %d\n", dt->tm_year);
		printf("rv5c387_convert_to_time: mon  = %d\n", dt->tm_mon);
		printf("rv5c387_convert_to_time: mday = %d\n", dt->tm_mday);
		printf("rv5c387_convert_to_time: hour = %d\n", dt->tm_hour);
		printf("rv5c387_convert_to_time: min  = %d\n", dt->tm_min);
		printf("rv5c387_convert_to_time: sec  = %d\n", dt->tm_sec);
	}
}

/*
 * Get the current time from the RTC
 */
void
rtc_get (struct rtc_time *tmp)
{
	unsigned char buf[RV5C387_RAM_SIZE];
	int ret;

	if (!setup_done)
		rv5c387_enable();

	if (!setup_done)
		return;

	memset(buf, 0, sizeof(buf));

	/* note that this returns reg. 15 in buf[0] */
	ret = rv5c387_readram(buf, RV5C387_RAM_SIZE);
	if (ret != 0) {
		printf("%s: failed\n", __FUNCTION__);
		return;
	}

	rv5c387_convert_to_time(tmp, buf);

	return;
}

/*
 * Set the RTC
 */
void
rtc_set (struct rtc_time *tmp)
{
	unsigned char buf[RV5C387_RAM_SIZE], reg_24h_mode;
	int ret;

	if (!setup_done)
		rv5c387_enable();

	if (!setup_done)
		return;

	if(rtc_debug > 2) {
		printf("rtc_set: tm_year = %d\n", tmp->tm_year);
		printf("rtc_set: tm_mon	 = %d\n", tmp->tm_mon);
		printf("rtc_set: tm_mday = %d\n", tmp->tm_mday);
		printf("rtc_set: tm_hour = %d\n", tmp->tm_hour);
		printf("rtc_set: tm_min	 = %d\n", tmp->tm_min);
		printf("rtc_set: tm_sec	 = %d\n", tmp->tm_sec);
	}

	memset(buf, 0, sizeof(buf));

	/* only read register containing 12/24hour mode */
	ret = i2c_read(CFG_I2C_RTC_ADDR, REGADR_24HOUR_MODE << REGADDR_SHIFT, REGADDR_LEN, &reg_24h_mode, 1);
	if (ret == 0) {
		buf[REGADR_SEC] = bin2bcd(tmp->tm_sec);
		buf[REGADR_MIN] = bin2bcd(tmp->tm_min);
		if (TWELVE_HOUR_MODE(reg_24h_mode))
		{	/* 12-hour-mode */
			if ( (tmp->tm_hour) > 12 )
			{	/* PM */
				buf[REGADR_HOUR] = 0x20 | bin2bcd((tmp->tm_hour) - 12);
			}
			else if( (tmp->tm_hour) == 12 )
			{	/* PM12 */
				buf[REGADR_HOUR] = 0x20 | bin2bcd(12);
			}
			else if( (tmp->tm_hour) == 0 )
			{	/* AM12 */
				buf[REGADR_HOUR] =        bin2bcd(12);
			}
			else
			{	/* AM */
				buf[REGADR_HOUR] =        bin2bcd(tmp->tm_hour);
			}
		}
		else
		{	/* 24-hour-mode */
			buf[REGADR_HOUR] = bin2bcd(tmp->tm_hour);
		}

		buf[REGADR_WDAY] = bin2bcd(tmp->tm_wday);
		buf[REGADR_DAY] = bin2bcd(tmp->tm_mday);
		buf[REGADR_CENT_MONTH] = bin2bcd(tmp->tm_mon);
		if (tmp->tm_year < 1900 || tmp->tm_year > 2099)
		{
			printf("WARNING: year should be between 1900 and 2099!\n");
		}
		if( tmp->tm_year >= 2000 )
		{
			buf[REGADR_CENT_MONTH] |= MASK_YEARBIT_19_20;
		}
		buf[REGADR_YEAR] = bin2bcd(tmp->tm_year % 100);

		ret = i2c_write(CFG_I2C_RTC_ADDR, REGADR_SEC << REGADDR_SHIFT, REGADDR_LEN, buf, (REGADR_YEAR - REGADR_SEC + 1));
		if (ret != 0)
		{
			printf("rv5c387_set_datetime(), i2c_master_send() returned %d\n",ret);
		}
	}

	return;
}

/*
 * Reset the RTC. We set the date back to 1970-01-01.
 */
void
rtc_reset (void)
{
	struct rtc_time tmp;
	unsigned char buf[RV5C387_RAM_SIZE];
	int ret;

	if (!setup_done)
		rv5c387_enable();

	if (!setup_done)
		return;

	tmp.tm_year = 1970;
	tmp.tm_mon = 1;
	/* Jan. 1, 1970 was a Thursday */
	tmp.tm_wday= 4;
	tmp.tm_mday= 1;
	tmp.tm_hour = 0;
	tmp.tm_min = 0;
	tmp.tm_sec = 0;

	buf[REGADR_CTRL1] = MASK_MODEBIT_12_24H;	/* Set only 24H mode bit */
	buf[REGADR_CTRL2] = 0;	/* Set default (with clearing XSTP) */
	ret = i2c_write(CFG_I2C_RTC_ADDR, REGADR_CTRL1 << REGADDR_SHIFT, REGADDR_LEN, &buf[REGADR_CTRL1], (REGADR_CTRL2 - REGADR_CTRL1 + 1));
	if (ret != 0)
	{
		printf("rv5c387_reset(), i2c_write() returned %d\n",ret);
		return;
	}

	rtc_set(&tmp);

	printf ("RTC:	%4d-%02d-%02d %2d:%02d:%02d UTC\n",
		tmp.tm_year, tmp.tm_mon, tmp.tm_mday,
		tmp.tm_hour, tmp.tm_min, tmp.tm_sec);

	return;
}

static unsigned int
bcd2bin (unsigned char n)
{
	return ((((n >> 4) & 0x0F) * 10) + (n & 0x0F));
}

static unsigned char
bin2bcd (unsigned int n)
{
	return (((n / 10) << 4) | (n % 10));
}
#endif /* defined(CONFIG_RTC_RV5C387A) && (CONFIG_COMMANDS & CFG_CMD_DATE) */
